package Sub::Junctive;
use Lingua::Wordnet;

use warnings;
use Carp;

our $VERSION = '0.01';

my $wn = Lingua::Wordnet->new;
$DB::single=1;
my $DEBUG;
my %classes;
my %syncache;
my %antcache;

sub import {
	my $caller = caller;
	*{$caller."::AUTOLOAD"} = mit($caller);
	$DEBUG = grep /-debug/, @_;
	$classes{$caller} = !grep /-runtime/, @_;
}

INIT {
	foreach my $class (keys %classes) {
		poena($class);
	}
}

sub poena {
	my ($class) = @_;
	delete $syncache{$class};
	foreach my $sub ( grep *{$class."::".$_}{CODE}, keys %{$class."::"}) {
		junctive($class, $sub);
		versive($class, $sub);
	}
}

sub mit {
	my $orig_AUTOLOAD = *{"$_[0]::AUTOLOAD"}{CODE};
	return sub : lvalue {
	    my ($caller, $method) = $AUTOLOAD =~ m{^(.*)::(.*)$};
	    poena($caller) unless $classes{$caller};
	    if (my @syns = keys %{$syncache{$caller}{$method}||{}}) {
		croak "Ambiguous call to $method (could be "
		    . join("() or ", @syns) . "())"
			if @syns > 1;
		goto &{$syns[0]}
	    }
	    if (my @ants = keys %{$antcache{$caller}{$method}||{}}) {
		return Sub::Junctive::Antonym->new(\@ants, $method, \@_)
		    if defined wantarray;
	    }
	    if ($orig_AUTOLOAD) { goto &$orig_AUTOLOAD }
	    croak "Can't find synonym for $method";
	}
}

sub sume {
	my ($method,$type,$relation) = @_;
	my @synsets = $wn->lookup_synset($method,$type) or return;
	if ($relation) {
		@synsets = map {$_->$relation} @synsets;
	}
	my @syns = map $_->words, @synsets;
	s/\W.*// foreach @syns;
	my %alts;
	@alts{@syns} = ();
	return keys %alts;
}

sub junctive {
	my ($caller, $method) = @_;
	my @words = (sume($method,'v'), sume($method,'n'),
		     sume($method,'a'), sume($method,'s'));
	# print "Could be:\n\t", join "\n\t", @$words, "\n";
	foreach $word (@words) {
		$syncache{$caller}{$word}{$caller."::".$method}++;
	}
}

sub versive {
	my ($caller, $method) = @_;
	my @words = (sume($method,'v','antonyms'), sume($method,'n','antonyms'),
		     sume($method,'a','antonyms'), sume($method,'s','antonyms'));
	# print "Could be:\n\t", join "\n\t", @$words, "\n";
	foreach $word (@words) {
		$antcache{$caller}{$word}{$caller."::".$method}++;
	}
}

package Sub::Junctive::Antonym;
use Carp;

sub new {
	my ($class, $antonyms, $method, $args) = @_;
	bless {ants=>$antonyms, meth=>$method, args=>$args}, $class;
}

use overload
	'!' => sub {
	    my ($self) = @_;
	    local *ants = $self->{ants};
	    croak "Ambiguous call to !$self->{meth} (could be "
		    . join("() or ", @ants) . "())"
			if @ants > 1;
	    $self->{negated}=!$self->{negated};
	    goto &{$ants[0]}
	},
	nomethod => 'AUTOLOAD';

sub AUTOLOAD {
	my ($self) = @_;
	croak "Can't find synonym for $self->{meth}"
		unless $self->{negated} || $self->{reported}++;
}

1;
__END__
